package exosearch;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import search.ExternalSearchV1;
import search.Heuristic;
import search.Searchable;
import search.WorkerListener;


/**
 * Simple AlphaBeta search implementation
 *
 * @author Stephen Tavener
 */
public final class Negamax implements ExternalSearchV1
{
	public static class TimeoutException extends RuntimeException { private static final long serialVersionUID = 1L; }
	
	public static class Position {

		public int heuristicValue;
		public int bestMove = -1;
		public int lastKiller = -1;
		public int value;
		
	}
	public static class Move {

		public int butterflyValue;
		public int histValue;
		
	}
	public static class Line {
		long[] moves = new long[0];
		public void merge(long move, Line line) 
		{
			//System.out.println("Merging "+move);
			moves = new long[line.moves.length+1];
			System.arraycopy(line.moves, 0, moves, 1, line.moves.length);
			moves[0] = move;
		}
	}
	
	private static final String SEARCH_DURATION_S = "searchDuration(s)";

	private static final int MAX_VAL = Integer.MAX_VALUE - 100; // Don't use min and max, because not safe to negate them
	private static final int MIN_VAL = -MAX_VAL;

	private static final int WIN_DILUTION_FACTOR = 100;
	
	private WorkerListener listener;
	private long startAt;
	private long endAt;
	private Map<Long,Position> positions = new HashMap<>();
	private Map<Long,Move> moves = new HashMap<>();

	private int activePlayer;

	private long lastReportTime;

	private int searchDepth;
	
	private List<Line>pvs = new ArrayList<>(); // principal variations
	Line pv; // Last pv
	
	public Negamax() { /* NO OP */ }

	@Override
	public double[] search(
			final WorkerListener searchListener, 
			final Map<String,Object> parameters,
			final Searchable masterGame)
	{	
		final Heuristic game = (Heuristic) masterGame.deepCopy();
		activePlayer = game.getOwner();
		lastReportTime = 0;
		positions.clear();
		moves.clear();
		
		listener = searchListener;
		
		final int n = game.numActions();

		final long searchTime = (long) (((Number)parameters.get(SEARCH_DURATION_S)).doubleValue() * 1000);
		
		startAt = System.currentTimeMillis();
		endAt  = startAt + searchTime;	
		
		//	Search...
		if (listener != null) listener.startWork((int) searchTime);

		searchDepth = 0;
		pv = null;
		
		while (true)
		{
			try 
			{				
				// Now perform the full search
				final long start = System.currentTimeMillis();
				long lastSearchT = 0;
				
				final Line line = new Line();
				
				final Heuristic state = game.deepCopy();
				final int score = negamax(state, searchDepth, MIN_VAL, MAX_VAL, line);
				lastSearchT = System.currentTimeMillis()-start;
				
				// NOTE: update PV only if we complete the pass - otherwise, we get much weaker, unfortunately :(
				pvs.add(line);
				pv = line;
				
				if (score < MIN_VAL/2) break; // solved - loss
				if (score > MAX_VAL/2) break; // solved - win
				if (System.currentTimeMillis()+lastSearchT > endAt) break; // No time for another pass, so why bother?
				searchDepth++;
			} 
			catch (@SuppressWarnings("unused") TimeoutException t) 
			{
				break;	// Search cancelled due to insufficient time
			}			
		}
		printPV ((Heuristic)masterGame, pv);
		
		final double[] results = new double[n];
		for (int mv = 0; mv < n; mv++) {
			final Position pos = positions.get(Long.valueOf(game.hash()^game.relativeActionHash(mv)));
			if (pos != null) results[mv] = Math.tanh(pos.value);
		}
		
		return results;
	}
	
	private void printPV(Heuristic master, Line line) 
	{
		final Heuristic game = master.deepCopy();
		
		StringBuilder sb = new StringBuilder();
		sb.append("PV (depth="+searchDepth+"): ").append(Arrays.toString(line.moves)).append("    ");
		
		for (int pos = 0; pos < line.moves.length; pos++) 
		{
			StringBuilder gameHashes = new StringBuilder();
			final long relHash = line.moves[pos];
			int mv = -1;
			for (int i = 0; i < game.numActions(); i++) 
			{
				gameHashes.append(",").append(game.relativeActionHash(i));
				if (game.relativeActionHash(i)==relHash) 
					mv = i; 
			}
			//System.out.println(gameHashes);
			if (mv == -1) break;
			if (sb.length() > 4) sb.append(",");
			sb.append(mv);
			game.applyNthAction(mv);
		}
		System.out.println(sb.toString());
	}

	private final int negamax (
			final Heuristic state,
			final int depth,
			int alpha, 
			int beta,
			final Line parentLine)
	{
		final long t = System.currentTimeMillis();
		if (t > endAt) throw new TimeoutException();
		if (listener!=null && t > lastReportTime+100) {
			lastReportTime = t;
			listener.reportProgress((int) (t-startAt));
			System.out.println(alpha+" v "+beta);
		}
		if (listener != null && listener.isStopRequested()) throw new TimeoutException();
		
		final Position entry = getOrCreatePosition(state.hash());
		if (state.isTerminal()) return entry.value = terminalState (state, searchDepth-depth);	// Solved? Return actual value, for speed		
		if (depth == 0) return entry.value = heuristic (state, entry);							// Maximum depth? Calculate value

		// Check best response from previous passes
		Line line = new Line();

//		final int bestResponse = entry.bestMove;
//		if (bestResponse >= 0 && bestResponse < state.numActions()) 
//		{
//			final Heuristic child = state.deepCopy();
//			child.applyNthAction(bestResponse);
//			final boolean samePlayer = child.getOwner()==state.getOwner();
//			final int value = negamax(
//					child,
//					depth-1,
//					samePlayer?alpha:-beta,
//					samePlayer?beta:-alpha, 
//					line);
//			if (value >= beta) return beta;
//			if (value > alpha) {
//				alpha = value;
//				entry.value = value;
//				parentLine.merge(state.hash()^child.hash(), line);
//			}
//		}
//
		final int[] order = getOrder (state, entry, depth);
		int value = MIN_VAL;

		// Loop through the rest of the moves
		for (int moveIndex : order) 
		{
//			if (moveIndex==bestResponse) continue;
			
			final Heuristic child = state.deepCopy();
			child.applyNthAction(moveIndex);
			final boolean samePlayer = child.getOwner()==state.getOwner();
			value = negamax(
					child,
					depth-1,
					samePlayer?alpha:-beta,
					samePlayer?beta:-alpha, 
					line);
			
			// Check for cutoffs, store values
			if (value >= beta)
			{
				// Mark butterflies - moves that did not cause a cutoff
				for (final int mv : order) 
				{
					if (mv == moveIndex) break;
					final long hash = state.relativeActionHash(mv);
					Move mvEntry = getOrCreateMove(hash);
					mvEntry.butterflyValue++;
				}
				// Mark cutoff
				final long hash = state.relativeActionHash(moveIndex);
				Move mvEntry = getOrCreateMove(hash);
				mvEntry.histValue++;

				entry.lastKiller = moveIndex;
				return value;
			}
			
			if (value > alpha)
			{
				alpha = value;
				entry.value = value;
				entry.bestMove = moveIndex;
				parentLine.merge(state.relativeActionHash(moveIndex), line);
			}
		}
		
		// Return best value (for some definition of best)
		entry.value = value;
		return value;
	}


	private Position getOrCreatePosition (final long hash) 
	{
		Position pos = positions.get(Long.valueOf(hash));
		if (pos == null) {
			pos = new Position();
			positions.put(Long.valueOf(hash), pos);
		}
		return pos;

	}

	private Move getOrCreateMove(final long hash) 
	{
		Move mvEntry = moves.get(Long.valueOf(hash));
		if (mvEntry == null) {
			mvEntry = new Move();
			moves.put(Long.valueOf(hash), mvEntry);
		}
		return mvEntry;
	}

	private final int[] getOrder(
			final Heuristic state, 
			final Position entry, 
			final int depth)
	{
		final int actionCount = state.numActions();
		final boolean[] isInList = new boolean[actionCount];
		final int[] results = new int[actionCount];
		int pos = 0;
	
		// Find suitable move order - may be a subset of the full list
		final int[] order = state.recommendedMoveOrder(activePlayer, searchDepth, searchDepth-depth, -1);

		// Remove moves that have already been considered...
		final int bestResponse = entry.bestMove;
		if (bestResponse >= 0 && bestResponse < actionCount) isInList[bestResponse] = true;
		
		// Add move(s) from pv
		// NOTE: this may not be in order, if the game is filtering
		if (pv != null && depth < pv.moves.length)
			for (int mv=0; mv<actionCount; mv++)
				if (!isInList[mv]) {
					final long mvHash = state.relativeActionHash(mv);
					if (pv.moves[depth] == mvHash) {
						results[pos++] = mv;
						isInList[mv] = true;	
					}
				}

		// Killer moves cause cutoffs
		{
			final int k = entry.lastKiller;
			if (k >= 0 && k < actionCount && !isInList[k]) {
				results[pos++] = k;
				isInList[k] = true;	
			}	
		}
		
		// Killer moves cause cutoffs
		{
			final int k = entry.bestMove;
			if (k >= 0 && k < actionCount && !isInList[k]) {
				results[pos++] = k;
				isInList[k] = true;	
			}	
		}
		
		// Look for butterflies (moves that are frequently skipped)
		for (int mv=0; mv<actionCount; mv++) 
			if (!isInList[mv]) {
				final long mvHash = state.relativeActionHash(mv);
				final Move mvEntry = getOrCreateMove(mvHash);
				if (mvEntry.butterflyValue < mvEntry.histValue) {
					results[pos++] = mv;
					isInList[mv] = true;	
				}
			}
		
		// Copy remainder of list...
		for (int mv : order) 
			if (!isInList[mv]) 
				results[pos++] = mv;
		
		if (pos == results.length) return results;
		return Arrays.copyOf(results, pos);
	}


	private static int heuristic (final Heuristic domState, final Position entry) 
	{
		if (entry.heuristicValue != 0) return entry.heuristicValue;
		
		final Heuristic d = domState.deepCopy();
		
		final int[] scores = d.calcHeuristics();
		int score = 0;
		for (int p=1; p<scores.length; p++)
			score += p==domState.getOwner() ? scores[p] : -scores[p];
		
		entry.heuristicValue = score;
		
		return score; 
	}
	
	private static int terminalState (final Heuristic domState, final int depth)
	{
		switch(domState.mcScore(domState.getOwner())) 
		{		
		case Searchable.MC_WIN: return MAX_VAL - depth*WIN_DILUTION_FACTOR;		
		case Searchable.MC_LOSS: return MIN_VAL + depth*WIN_DILUTION_FACTOR;
		case Searchable.MC_DRAW: return 0;
		default: return 0;		// Anything else is weird - treat as a draw for now
		}
	}

	@Override
	public Map<String, Object> parameters() 
	{
		final Map<String, Object> parameters = new HashMap<>();
		parameters.put(SEARCH_DURATION_S, Double.valueOf(1));
		return parameters;
	}
	
	@Override
	public String toString() {
		return "AB(Negamax)";
	}}
