package exosearch;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ThreadLocalRandom;

import search.ExternalSearchV1;
import search.Searchable;
import search.WorkerListener;

/**
 * @author Stephen Tavener
 */
public class UCT implements ExternalSearchV1 
{
	private static final float c = (float) Math.sqrt(2);
	private static final String SEARCH_DURATION_S = "searchDuration(s)";

	public static class Node 
	{
		final Node parent;
		final int scoreOwner;
		final int index;
		final int numChildren;
		int visits;
		int sum;
		int actualChildren;
		private Node[] children; // most entries in tree will be leaf nodes, don't allocate unless needed!

		public Node (final Node parent, final int scoreOwner, final int index, final int numChildren) {
			this.parent = parent;
			this.scoreOwner = scoreOwner;
			this.index = index;
			this.numChildren = numChildren;
		}

		public boolean allVisited() { return children != null && actualChildren == children.length; }
		
		public Node nthChild (final int n) { 
			if (children == null) children = new Node[numChildren];
			return children[n];
		}
		
		public void addNthChild (final int n, final Node child) 
		{ 
			if (children == null) children = new Node[numChildren];
			if (children[n]==null) actualChildren++;
			children[n] = child;
		}
	}
	
	public UCT() { /* NO OP */ }

	@Override
	public double[] search (final WorkerListener listener, final Map<String,Object> parameters, final Searchable game)
	{	
		final Searchable gameCopy = game.deepCopy();
		final int n = gameCopy.numActions();

		final long searchTime = (long) (((Number)parameters.get(SEARCH_DURATION_S)).doubleValue() * 1000);
		
		final long startAt = System.currentTimeMillis();
		final long endAt  = startAt + searchTime;	

		/** Start of search - moves from current board state */
		final Node rootNode = new Node(null, game.getOwner(), -1, n);		

		//	Search...
		if (listener != null) listener.startWork((int) searchTime);
	
		int iterations = 0;
		while (System.currentTimeMillis() <= endAt)  
		{
			final Searchable state = game.deepCopy();
			state.reify();
			
			final Node twig = descendTree(state, rootNode);		// 1. Select best leaf node (may be the root)
			//if (twig == rootNode) {
			//	System.out.println("Exploring root");
			//}
			// 2. Expand to new leaf, unless this node is solved or has been visited less than the requisite number of times
			final Node leaf = expandTree(state, twig);

			// 3. perform simulations, propagate results
			// Note that the simulation may well trash the domain state, so may need to take copy...
			runSimulation(state);								// 3a. Run simulation (unless this is an end node)
			backpropagate(leaf, state);							// 3b. Pass the result back up the tree					

			// 4. send feelgood message back to user
			if ((++iterations&0b111111)==0 && listener != null) listener.reportProgress((int) (endAt-System.currentTimeMillis()));;
			if (listener != null && listener.isStopRequested()) break;
		}	
		
		final double[] results = new double[n];
		for (int i = 0; i < n; i++) 
		{
			Node node = rootNode.nthChild(i);
			if (node==null||node.visits==0) continue;
			results[i] = node.sum / (double) Math.max(1, node.visits);
		}
		
		System.out.println("UCT: "+iterations+" its");
		return results;
	}
	
	private static final Node descendTree(final Searchable domState, final Node node)
	{
		Node current = node;
		while (current.allVisited() && !domState.isTerminal())
		{
			final Node next = UCBChoice(current);
			if (next == null) return current;

			current = next;
			domState.applyNthAction(current.index);
		} 	
		return current;
	}

	private static final Node expandTree(final Searchable state, final Node parent)
	{
		// Safety guard - may get here if tree is fully explored
		if (state.isTerminal()) return parent;
		
		//	Get random move index from unused children
		final int moveIndex = newChildIndex(parent);
		
		// Save owner because all child nodes are owned by it.  Note the disparity between node owner and game state owner!
		final int childOwner = state.getOwner();
		
		// Step into node to check for terminus and calculate number of actions 
		state.applyNthAction(moveIndex);

		// Set up the child node
		final Node child = new Node(parent, childOwner, moveIndex, state.numActions());
		parent.addNthChild(moveIndex, child);
		
		return child;
	}
	
	private final static int newChildIndex (final Node parent) 
	{
		final int numActions = parent.numChildren;
		final int start = ThreadLocalRandom.current().nextInt(numActions);
		int idx = start;

		while (parent.nthChild(idx)!=null)
		{
			idx++; 
			if (idx >= numActions) idx = 0;
			if (idx == start) throw new IllegalStateException("No unused child found!");
		}
		return idx;
	}

	private static final void runSimulation(final Searchable state)
	{
		state.evaluate();
	}

	private static final void backpropagate(final Node node, final Searchable state)
	{
		for (Node current = node; current != null; current = current.parent)
		{
			current.visits++;
			current.sum += state.mcScore(current.scoreOwner);
		}
	}

	private static final Node UCBChoice(final Node parent)
	{
		// generic speedup
		final int childCount = parent.numChildren; 
		if (childCount==1) return parent.nthChild(0);
		
		Node bestNode = null;
		double bestVal  = Float.NEGATIVE_INFINITY; 
		
		// Exploration value is: value + const * sqrt(log(parentVists)/childVisits) - separate out the constant part...
		final double exploreMultiplier = Math.sqrt(Math.log(parent.visits)) * c;
		
		// Loop through nodes, find best one.
		for (int i=0; i < childCount; i++)
		{
			final Node ptr = parent.nthChild(i);
			
			//	Determine exploitation component (win rate)
			final float exploit = ptr.sum / (float)ptr.visits;

			//	Determine exploration component
			final double explore = exploreMultiplier / Math.sqrt(ptr.visits);

			//	Combine the exploration and exploitation estimates
			final double value = exploit + explore;
			
			if (value > bestVal)
			{
				bestVal  = value;
				bestNode = ptr;
			} else if (value == bestVal && ThreadLocalRandom.current().nextInt(childCount)==0) {
				bestNode = ptr;				
			}
		}

		return bestNode;
	}


	@Override
	public Map<String, Object> parameters() 
	{
		final Map<String, Object> parameters = new HashMap<>();
		parameters.put(SEARCH_DURATION_S, Double.valueOf(1));
		return parameters;
	}
	
	@Override
	public String toString() {
		return "UCT(Vanilla)";
	}
}