package external;

import java.util.Arrays;
import java.util.Map;
import java.util.Random;

/**
 * 
 * @author Stephen Tavener
 */
public final class BreakthroughExt implements ExternalGameV2
{
	/* ==================================================================================================================================
	 * CONSTANTS
	 * ================================================================================================================================== */
	private static final String[] COLNAMES = { "a", "b", "c", "d", "e", "f", "g", "h", "i", "j" };
	private static final String[] ROWNAMES = { "1", "2", "3", "4", "5", "6", "7", "8", "9", "10" };

	private static final int MAX_PLAYERS = 2;
	private static final int MAX_ROWS = 10;
	private static final int MAX_COLS = 10;
	
	private static final int WHITE = 1;
	private static final int BLACK = 2;
	private static final int EMPTY = 0;

	/* ==================================================================================================================================
	 * CONSTANTS (Zobrist hashes)
	 * 
	 * Give or take the occasional collision:
	 * - each position will be indicated by a unique hash
	 * - a move can be identified by the difference between the before/after hashes
	 * This allows the AI to reuse values learned in one branch of the tree elsewhere
	 * 
	 * Note: Use a constant seed. Ai Ai will at some point start using the hashes in opening library and endgame library generation
	 *       In this event, it will be important that the same hashes are generated between runs 
	 * ================================================================================================================================== */
	private static final long[] PHASHES = new long[MAX_PLAYERS+1];
	private static final long[][][] BHASHES = new long[MAX_PLAYERS+1][MAX_ROWS][MAX_COLS];

	static {
		final Random r = new Random(-5514723336353821969L);
		
		for (int i = 0; i < BHASHES.length; i++)
			for (int j = 0; j < BHASHES[0].length; j++)
				for (int k = 0; k < BHASHES[0][0].length; k++)
					BHASHES[i][j][k] = r.nextLong();
		
		for (int i = 0; i < PHASHES.length; i++)
			PHASHES[i] = r.nextLong();
	}
			
	/*
	 * Fields can be divided into three types:
	 * - state: these are mutable used calculating the list of moves available; they must be deep copied and included in the hash
	 * - reference data: initialised once and never changed; these can be copied by reference 
	 * - derived data: this may be deep copied but could be discarded and regenerated as needed, so does not contribute to the hash
	 */

	/* ===================================================================================================================================
	 * REFERENCE
	 * =================================================================================================================================== */
	private int rows;
	private int cols;

	private int pieceWt;
	private int rankWt;
	private int safeWt;
	private int unblockedWt;
	
	/* ===================================================================================================================================
	 * STATE
	 * =================================================================================================================================== */
	private int player;
	private int[][] board; // board[row][col]
	
	/* ===================================================================================================================================
	 * DERIVED
	 * =================================================================================================================================== */
	private long hash;
	private int winner;
	/** In this game, each move can be simply described by from and to coordinates. */
	private int[] moveList;
	private int moveCount;
	
	
	/** 
	 * Default constructor (all the action is in init)!
	 */
	public BreakthroughExt() {	/* NO OP - see init */ }
	
	/** 
	 * Copy constructor - used by deepCopy() 
	 * @param that 
	 */
	private BreakthroughExt(final BreakthroughExt that) 
	{	
		this.moveList = that.moveList.clone();
		this.moveCount = that.moveCount;
		this.winner = that.winner;
		this.hash = that.hash;

		this.rows = that.rows;
		this.cols = that.cols;
		
		this.player = that.player;
		
		this.board = new int[that.board.length][];
		for (int r = 0; r < that.board.length; r++)
			this.board[r] = that.board[r].clone();
		
		this.pieceWt = that.pieceWt;
		this.rankWt = that.rankWt;
		this.safeWt = that.safeWt;
		this.unblockedWt = that.unblockedWt;

	}

	@Override public final ExternalGameV2 deepCopy() { return new BreakthroughExt(this); }

	@SuppressWarnings("unchecked")
	@Override
	public final void init(Map<String, Object> json) 
	{
		// Here, we are getting the board size (rows, cols) from the .mgl file
		// Note that it's buried pretty deep at world.board.topology.rows, etc.
		final Map<String, Object> worldJson = (Map<String, Object>) json.get("world");
		final Map<String, Object> boardJson = (Map<String, Object>) worldJson.get("board");
		final Map<String, Object> topologyJson = (Map<String, Object>) boardJson.get("topology");
		
		rows = ((Number) topologyJson.get("rows")).intValue();
		cols = ((Number) topologyJson.get("cols")).intValue();
		if (rows > 10 || rows < 4 || cols > 10 || cols < 2) throw new IllegalArgumentException("Board size must be 4-10 rows x 2-10 columns!");
		
		// Get the heuristics values in the same way, allowing for external tuning
		final Map<String, Object> heuristics = (Map<String, Object>) json.get("heuristics");
		pieceWt = ((Number) heuristics.get("pieceWt")).intValue();
		rankWt = ((Number) heuristics.get("rankWt")).intValue();
		safeWt = ((Number) heuristics.get("safeWt")).intValue();
		unblockedWt = ((Number) heuristics.get("unblockedWt")).intValue();

		
		final int numPieces = 4*rows;		
		moveList = new int[numPieces * 3 * 4]; // maximum move list size - total pieces x 3 moves per piece x 4 coords: (r,c) to (r,c)
		moveCount = 0;
		winner = 0;
		
		// Initialise the board - a piece in each back of the two back rows
		board = new int[rows][cols];
		for (int c = 0; c < cols; c++) {
			board[0][c] = WHITE;
			board[1][c] = WHITE;
			board[rows-1][c] = BLACK;
			board[rows-2][c] = BLACK;
		}
		
		player = WHITE;
	}

	@Override
	public final int numMoves() 
	{
		// Generate the moves (no cost if already generated)
		generate();
		return Math.max(1, moveCount); // Always return at least one move; Ai Ai doesn't like empty move lists
	}

	private final void generate() 
	{
		// This is an optimisation; if the moves are already generated, don't do anything
		// moveCount will be cleared when the state changes
		if (moveCount != 0) return; // don't do unnecessary work
		
		// For each row
		//   For each column
		//      Is there a piece of my colour?
		//        Generate the moves for that piece; 
		//        - forward if empty
		//        - diagonally forward if empty or occupied by enemy; i.e. not a friend
		for (int r = 0; r < rows; r++)
			for (int c = 0; c < cols; c++)
				if (player == board[r][c])
				{
					// NOTE: there is always a row in front of every piece, or someone has won and we won't be generating moves any more!
					final int nextRank = (player == WHITE) ? r+1 : r-1;
					if (board[nextRank][c] == EMPTY) addMove(r, c, nextRank, c);
					if (c > 0 && board[nextRank][c-1] != player) addMove(r, c, nextRank, c-1);
					if (c+1 < cols && board[nextRank][c+1] != player) addMove(r, c, nextRank, c+1);
				}
	}

	private final void addMove(int rStart, int cStart, int rEnd, int cEnd) 
	{
		// Add a move to the move list - each move takes four cells; row and column for the start, row and column for the end
		// Note we don't need to add anything for captures - we can work that out when we apply the move
		int pos = moveCount*4;
		moveList[pos++] = rStart;
		moveList[pos++] = cStart;
		moveList[pos++] = rEnd;
		moveList[pos++] = cEnd;
		moveCount++;
	}

	@Override
	public final void applyNthMove(int n) 
	{
		generate();
		
		if (moveCount == 0) {	// If there are no moves possible, abandon the game
			winner = -1;
			return;
		}
		
		// Retrieve the move
		int pos = n*4;
		final int r0 = moveList[pos++];
		final int c0 = moveList[pos++];
		final int r1 = moveList[pos++];
		final int c1 = moveList[pos++];
		
		// Pick up the piece in the start cell
		// Remove any piece in the destination
		// Put the moving piece down in the destination 
		final int startOwner = pickUp(r0,c0);
		pickUp(r1,c1);
		putDown (startOwner, r1, c1);

		// Did we reach the back rank?  We won!
		if (player == WHITE && r1 == rows-1) winner = WHITE;
		if (player == BLACK && r1 == 0) winner = BLACK;
		
		// Change to next player
		changePlayer();
		
		moveCount = 0;
	}

	private final void changePlayer() 
	{
		// Note the hash does change here, appearances to the contrary:
		// - remove the old player hash,
		// - change the player
		// - apply the new player hash
		
		hash ^= PHASHES[player];	// remove old player hash
		player = player^3; // Magic : WHITE = 1 = binary 01, BLACK = 2 = binary 10; WHITE xor 3 = BLACK and vice versa
		hash ^= PHASHES[player];	// apply new player hash
	}

	private final int pickUp (final int r, final int c)
	{
		final int value = board[r][c];
		if (value != EMPTY) hash ^= BHASHES[value][r][c]; // Remove the hash IF there's something to pick up
		board[r][c] = EMPTY;
		return value;
	}

	private final void putDown (final int value, final int r, final int c)
	{
		board[r][c] = value; 
		hash ^= BHASHES[value][r][c]; // Add the hash for this piece
	}
	
	@Override
	public final int getOwner() {
		return player;
	}

	@Override
	public final boolean isTerminal() {
		return winner != 0;
	}

	@Override
	public final int mcScore(int plyr) {
		if (winner == plyr) return 1;
		if (winner == (plyr^3)) return -1;
		if (winner == -1) return 0;
		return 0;
	}

	@Override
	public final long hash(int moveIndex) {
		return hash;
	}

	@Override
	public final boolean isHeuristic() {
		return true;
	}

	@Override
	public final int[] calcHeuristics() 
	{
		// As we did with the hashes, it would be more efficient if the heuristics were calculated incrementally as the moves are applied
		// Done this way for clarity only
		
		// Ai Ai expects an array of values, one value for each player; note cell zero is not used
		final int[] scores = new int[MAX_PLAYERS+1];

		for (int r = 0; r < rows; r++)
			for (int c = 0; c < cols; c++) 
			{
				final int pieceOwner = board[r][c];
				if (pieceOwner == EMPTY) continue;

				final int adjR = (pieceOwner == WHITE) ? r : rows-1-r;
				
				// Pieces further ahead are more valuable - reward distance using triangle numbers
				int rankValue = (adjR+2) * (adjR+1) / 2;
				
				if (adjR == 0) {
					// Pieces on the back rank are valuable blockers; encourage the AI to leave a guard in the pattern - W W - - W W - etc */ 
					final boolean strong = (c&1) != ((c>>1)&1);
					rankValue = strong ? 15 : 5;
				} else if (c==0 || c==cols-1) {
					// Pieces on the edges are easier to block than centre pieces, so treat them as one rank behind where they really are */
					rankValue -= adjR+1;
				}
				
				scores[pieceOwner] += pieceWt;				// value of a piece, regardless of position
				scores[pieceOwner] += rankWt * rankValue;	// pieces further ahead on the board are more valuable

				// If piece cannot be captured, give it a bonus
				if (isSafe (r, c)) scores[pieceOwner] += safeWt * rankValue;
				
				// If piece is not blocked in the next two ranks (at most one piece in the code of movement), give it a bonus
				if (unblocked (r, c)) scores[pieceOwner] += unblockedWt * rankValue;
			}

		return scores;
	}

	/**
	 * @param r row
	 * @param c column
	 * @return true if the cone of movement in front of this piece contains no more than one piece
	 */
	private final boolean unblocked(int r, int c) 
	{
		final int enemy = board[r][c]^3;
		int count = 0;
		
		final int rp1 = (player == WHITE) ? r+1 : r-1;
		if (rp1 < 0 || rp1 >= rows) return true; // last rank!

		if (board[rp1][c] == enemy) count++;
		if (c > 0 && board[rp1][c-1] == enemy) count++;
		if (c+1 < cols && board[rp1][c+1] == enemy) count++;		
		if (count > 1) return false;
		
		final int rp2 = (player == WHITE) ? r+2 : r-2;
		if (rp2 < 0 || rp2 >= rows) return true;
		
		if (c > 1 && board[rp2][c-2] == enemy) count++;
		if (c > 0 && board[rp2][c-1] == enemy) count++;
		if (board[rp2][c] == enemy) count++;
		if (c+1 < cols && board[rp2][c+1] == enemy) count++;		
		if (c+2 < cols && board[rp2][c+2] == enemy) count++;		

		return (count < 2);
	}

	/**
	 * @param r row
	 * @param c column
	 * @return true if this piece is safe from capture
	 */
	private final boolean isSafe(int r, int c) 
	{
		final int enemy = board[r][c]^3;
		final int nextRank = (player == WHITE) ? r+1 : r-1;
		if (nextRank < 0 || nextRank >= rows) return true; // last rank!
		
		if (c > 0 && board[nextRank][c-1] == enemy) return false;
		if (c+1 < cols && board[nextRank][c+1] == enemy) return false;
		
		return true;
	}

	@Override
	public final void nullMove() {
		changePlayer();
	}

	@Override
	public final boolean isStable() {
		return true;
	}

	@Override
	public final int[] recommendedMoveOrder(int searchPlayer, int maxDepth, int currentDepth, int numVisits) 
	{
		// rule of thumb for AB; 
		// - worst ordering gives depth n
		// - random ordering gives depth sqrt(2).n 
		// - best ordering gives depth 2n
		// by default Ai A uses random ordering together with some enhancements which prioritise best moves from elsewhere in the tree
		// this gives something close to optimal ordering
		
		
		return null; // let the caller take care of this
	}

	@Override
	public final String[] drawables() {
		final String[] moves = new String[4*MAX_COLS];	// Maximum possible pieces
		
		int count = 0;

		for (int r = 0; r < rows; r++)
			for (int c = 0; c < cols; c++)
				switch (board[r][c]) {
				case EMPTY: 
					break; // nothing to do
				case WHITE: 
				case BLACK: 
					moves[count++] = "disc"+"#"+playerName(board[r][c])+"#"+positionName(r,c);
					break;
				}
		return Arrays.copyOf(moves, count);
	}

	@Override
	public final String actionDescription(int n) {
		generate();
		if (moveCount==0) return "Pass";
		
		int pos = n*4;
		final int r0 = moveList[pos++];
		final int c0 = moveList[pos++];
		final int r1 = moveList[pos++];
		final int c1 = moveList[pos++];

		return positionName(r0,c0)+"-"+positionName(r1,c1);
	}

	private static final String positionName (final int r, final int c) 
	{
		return COLNAMES[c]+ROWNAMES[r];
	}

	private static final String playerName (final int plyr) 
	{
		return plyr == WHITE ? "White" : "Black";
	}
	
	@Override
	public final String stateDescription() {
		if (winner != 0) {
			if (winner == -1) return "Draw";
			return playerName(winner)+" won";
		}
		return playerName(player) + " to move";
	}

}
