package search;

/**
 * Heuristics - optional interface for some games to improve play
 * Games implementing Heuristics should also:
 * @Override
	public boolean isHeuristic() {
		return true;
	} 
 * @author Stephen Tavener
 *
 */
public interface Heuristic extends Searchable {
	/**
	 * Calculates relative scores for each player.
	 * AB search will use the difference between the scores, but UCT implementations will return win/loss based on (player score)/(total score).
	 * To achieve this, each score must be >= 0, and total score must be greater than zero
	 * Players/Scores are 1-based, so entry 1 is the score for player 1, etc.
	 * @return array of scores for each player
	 */
	public int[] calcHeuristics();
	
	/**
	 * Used with quiescence search.
	 * True if the position is stable, and therefore heuristics are likely to be reliable
	 * Example; in chess stable usually implies no captures are possible.
	 * Default should be true - if not, the search may not terminate!
	 * Note that outside of Chess and Emergo, I haven't found many games where quiescence improves search
	 * @return true if position is stable
	 */
	public boolean isStable();
	
	@Override
	Heuristic deepCopy();
}
