package search;


/**
 * The Searchable interface is a contract between the domain object (e.g. the current position in a game of chess) and the Search engine. 
 * All games implement this interface. 
 * 
 * @author Stephen Tavener
 */
public interface Searchable
{
	/** Win value */
	public static final int MC_WIN  =  1;
	/** Loss value */
	public static final int MC_LOSS = -1;
	/** Draw value */
	public static final int MC_DRAW =  0;

	/**
	 * @return number of agents, ... from 1 upwards.  Implies whether the game is adversarial or not
	 */
	int numAgents();

	/**
	 * @return the owner of the current state (e.g. the player to move). First player has value 1, last player will have value equal to numAgents()
	 */
	int getOwner();

	/**
	 * @return number of available actions from this state.  If the game is not in a terminal state, there must always be one action, even if it is only "Pass".
	 */
	int numActions();

	/**
	 * Changes the current state to a new state by applying the nth action 
	 * @param n
	 */
	void applyNthAction(final int n);
	
	/**
	 * Checks if the game is in a terminal state
	 * NOTE: this may require some processing, so is not to be called unnecessarily.
	 * @return true if this is an end state
	 */
	boolean isTerminal();

	/**
	 * Called to obtain the score for a specific player - note that multi-player games can have multiple winners, losers, etc.
	 * If the game is not in a terminal state, behaviour is undetermined.
	 * Valid return values are MC_WIN, MC_DRAW, MC_LOSS
	 * @param player 
	 * @return returns the score for this player
	 */
	int mcScore (final int player);

	/**
	 * Called before a *full* search to make any one-off optimisations, etc.
	 * The idea here is to perform heuristic measurements, etc. that are too expensive to perform on a per-node basis
	 * As the search deepens, they will become less accurate, so there is a compromise between speed and accuracy as usual
	 * Determinisation should be done in reify() if at all!
	 *  
	 * NOTE: it is expected that this copy of the domain will be disposed of after the search, so state change is allowed (if inadvisable!) 
	 */
	void prepareForSearch();

	/**
	 * Determinisation - collapses a superposition of unknown data down to known values for the purpose of a search
	 * Used by Monte Carlo searches, but should not be used by MiniMax searches
	 * 
	 * NOTE: it is expected that this copy of the domain will be disposed of after the search, so state change is allowed (and probably required) 
	 */
	void reify();

	/**
	 * Evaluates the game.
	 * After this function call, the game should be in a terminal state, and mcScore should return the final score for the specified player
	 */
	void evaluate();
	
	/**
	 * Implementing classes should order moves according to most likely strength.  
	 * Some AIs can then use these during search to improve cutoffs.
	 * It is not necessary to return all moves; but it is necessary to return at least one move.
	 * 
	 * Example: most placement games should order moves from centre outwards, 
	 * Capture games should put captures first, starting with the most valuable.
	 * 
	 * The AI is free to re-order this list based on extra heuristics like killer moves.
	 * 
	 * @param searchPlayer Player who is searching
	 * @param maxDepth Maximum search depth for this pass. -1 if called outside of a search pass
	 * @param currentDepth Current search depth from root for this pass. -1 if called outside of a search pass
	 * @param numVisits UCT based searches only; number of times this node has been visited.  -1 for other searches
	 * 
	 * @return subset of available move indices in recommended order of preference
	 */
	public int[] recommendedMoveOrder(
			int searchPlayer,
			int maxDepth, 
			int currentDepth, 
			int numVisits);

	/**
	 * Board hashes are used by the AI to record positions for openings, endgames, and transposition tables
	 * Ideally:
	 * - two different positions will have different hashes
	 * - two identical positions will have the same hash
	 * 
	 * @return the current board hash
	 */
	long hash();
	
	/**
	 * Board hashes are used by the AI to record positions for openings, endgames, and transposition tables
	 * Ideally:
	 * - two different positions will have different hashes
	 * - two identical positions will have the same hash
	 * An action hash is the delta between two positions when the move is applied
	 * 
	 * @param index 
	 * @return the hash difference between before and after this move is applied
	 */
	long relativeActionHash(int index);
	
	/**
	 * Board hashes are used by the AI to record positions for openings, endgames, and transposition tables
	 * Ideally:
	 * - two different positions will have different hashes
	 * - two identical positions will have the same hash
	 * An action hash is the delta between two positions when the move is applied
	 * 
	 * @return the hash difference between last position and this position
	 */
	long lastHashDelta();
	
	/**
	 * @return true if this game provides heuristic information (and is therefore suitable for alpha-beta search).
	 * Cast the game to Heuristic to get the additional information.
	 */
	boolean isHeuristic();	
	
	/**
	 * Produces a deep copy of this object - necessary, because Java returns a shallow copy by default, 
	 * which can lead to corruption of the state during search.  
	 * 
	 * Some care is required here.  If you have an object O that is referred to from two board positions, 
	 * for example, you need to ensure that it is only copied once; otherwise references which were formerly 
	 * referring to the same object will be pointing to different objects
	 * 
	 * @return new game which is a perfect copy of this game
	 */
	Searchable deepCopy();
}
